%% load TCAD impedance

clc
clear all

impedancedata = readmatrix('TCAD_75C_1sun_MPP.csv');

%set values
area = 153; %cm2
L_set = 60; %in nH
Rs_set = 5; %in mOhm 

%load admittance data
frequency = impedancedata(:,1); % in Hz
omega = 2*pi*frequency;
admittance_G = impedancedata(:,2); %for device of 1 cm2
admittance_B = impedancedata(:,3); %for device of 1 cm2

%convert admittance data to impedance
impedance_R = (1/area)*(admittance_G./(admittance_G.^2+admittance_B.^2));
impedance_X = (1/area)*(-admittance_B./(admittance_G.^2+admittance_B.^2));
impedance_Mag = sqrt(impedance_R.^2+impedance_X.^2);
impedance_phase = atand(impedance_X./impedance_R);

%select frequency range
f_min = min(frequency);     %min frequency (Hz)
f_max = 1*10^6;             %max frequency (Hz)

[valuemin,indexmin]=min(abs(f_min-frequency)); f_min = frequency(indexmin);
[valuemax,indexmax]=min(abs(f_max-frequency)); f_max = frequency(indexmax);
frequency = frequency(indexmin:indexmax); omega = omega(indexmin:indexmax);
impedance_Mag = impedance_Mag(indexmin:indexmax); impedance_phase = impedance_phase(indexmin:indexmax); impedance_R = impedance_R(indexmin:indexmax); impedance_X = impedance_X(indexmin:indexmax);

%add series resistance and inductance
impedance_R = impedance_R+(Rs_set/10^3);
impedance_X = impedance_X+omega.*(L_set/10^9);
impedance_Mag = sqrt(impedance_R.^2+impedance_X.^2);
impedance_phase = atand(impedance_X./impedance_R);



%% global fitting for initial values
% impedance model contains a PN junction, series resistance, and inductor

% set initial value, lower bound, and upper bound for all the fitting parameters 

% PN junction resistance
RD_init = impedance_Mag(1);
RD_low = 0; %Ohm, lower bound
RD_high = Inf; %Ohm, higher bound

% Series resistance
Rser = Rs_set; %Rser_LOANA; %mOhm, fill in value from LOANA
Rser_low = Rser*0.5; %mOhm, lower bound
Rser_high = Rser*2; %mOhm, upper bound

% PN junction capacitance 
C_init = 10; % in uF 
C_low = 0; % in uF, lower bound
C_high = Inf; % in uF, upper bound

%inductance 
L_init = L_set; %L_init_set ; %nH
L_low = 0.8*L_set ; % in nH, lower bound
L_high = L_set*1.2; %in nH, upper bound

omegafit = [omega omega];
impedancefit = [impedance_R impedance_X];

fun1 = @(fit,omegafit) fit(1) + fit(2)./(1+(fit(2)^2)*(omegafit.^2).*(fit(3).^2));
fun2 = @(fit,omegafit) omegafit.*fit(4)-((fit(2)^2)*omegafit.*fit(3))./(1+(fit(2)^2)*(omegafit.^2).*(fit(3).^2));

fun = @(fit,omegafit) [fun1(fit,omegafit(:,1)),  fun2(fit,omegafit(:,2))];

fit0 = [Rser/1000 RD_init C_init*10^-6 L_init*10^-9];
lb = [Rser_low/1000 RD_low C_low*10^-6 L_low*10^-9];
ub = [Rser_high/1000 RD_high C_high*10^-6 L_high*10^-9];

fit = lsqcurvefit(fun,fit0,omegafit,impedancefit,lb,ub);
Rs_global = fit(1); RD_global = fit(2); C_global = fit(3); L_global = fit(4);

%% Fitting with final model including LH junction
% impedance model contains a PN junction, LH junction, series resistance, and inductor

%R_LH upper and lower bounds
R_LH_init = 0;    % in Ohm, initial  
R_LH_low = 0;            % in Ohm, lower bound
R_LH_high = 0;            % in Ohm, upper bound

%C_LH upper and lower bounds
C_LH_init = 0;        % in F, initial  
C_LH_low = 0;         % in F, lower bound
C_LH_high = 0; % in F, upper bound

omegafit = [omega omega];
impedancefit = [impedance_R impedance_X];

fun1 = @(fit,omegafit) fit(1)+ (fit(2)./(1+(fit(2)^2)*(omegafit.^2).*(fit(3).^2))) + (fit(4)./(1+(fit(4)^2)*(omegafit.^2).*(fit(5).^2)));
fun2 = @(fit,omegafit) omegafit.*fit(6)-((fit(2)^2)*omegafit.*fit(3))./(1+(fit(2)^2)*(omegafit.^2).*(fit(3).^2))-((fit(4)^2)*omegafit.*fit(5))./(1+(fit(4)^2)*(omegafit.^2).*(fit(5).^2));

fun = @(fit,omegafit) [fun1(fit,omegafit(:,1)),  fun2(fit,omegafit(:,2))];

fit0 = [Rser/1000 RD_global C_global R_LH_init C_LH_init L_init*10^-9];
lb = [Rser_low/1000 RD_low C_low*10^-6 R_LH_low*10^-3 C_LH_low*10^-6 L_low*10^-9];
ub = [Rser_high/1000 RD_high C_high*10^-6 R_LH_high C_LH_high L_high*10^-9];

fit = lsqcurvefit(fun,fit0,omegafit,impedancefit,lb,ub);
Rs_fit = fit(1); Rj_fit = fit(2); Cj_fit = fit(3); R_LH_fit = fit(4); C_LH_fit = fit(5); L_fit = fit(6); 

disp(['CNLS best-fit values are Cj = ',num2str(Cj_fit*10^6/area),' uF/cm^2, ', ...
    'Rj = ',num2str(Rj_fit*area),' Ohm*cm^2, ',...
    'Rs = ',num2str(Rs_fit*10^3),' mOhm, ',...
    'L = ',num2str(L_fit*10^9),' nH ']);

%% Fitting quality assessment

Real_fit_full = fit(1)+ (fit(2)./(1+(fit(2)^2)*(omega.^2).*(fit(3).^2))) + (fit(4)./(1+(fit(4)^2)*(omega.^2).*(fit(5).^2)));
Compl_fit_full = omega.*fit(6)-((fit(2)^2)*omega.*fit(3))./(1+(fit(2)^2)*(omega.^2).*(fit(3).^2))-((fit(4)^2)*omega.*fit(5))./(1+(fit(4)^2)*(omega.^2).*(fit(5).^2)); 
Mag_fit_full = sqrt(Real_fit_full.^2+Compl_fit_full.^2);
Angle_fit_full = atand(Compl_fit_full./Real_fit_full);

%calculate NRMSE magnitude
MSE_mag = goodnessOfFit(impedance_Mag,Mag_fit_full,'MSE');
RMSE_mag = sqrt(MSE_mag);
NRMSE_mag = abs(RMSE_mag/(max(impedance_Mag)-min(impedance_Mag)));

%calculate RMSE phase
MSE_phase = goodnessOfFit(impedance_phase,Angle_fit_full,'MSE');
RMSE_phase = sqrt(MSE_phase);
NRMSE_phase = abs(RMSE_phase/(max(impedance_phase)-min(impedance_phase)));

%calculate NRMSE resistance
MSE_real = goodnessOfFit(impedance_R,Real_fit_full,'MSE');
RMSE_real = sqrt(MSE_real);
NRMSE_real = abs(RMSE_real/(max(impedance_R)-min(impedance_R)));

%calculate NRMSE reactance
MSE_complex = goodnessOfFit(impedance_X,Compl_fit_full,'MSE');
RMSE_complex = sqrt(MSE_complex);
NRMSE_complex = abs(RMSE_complex/(max(impedance_X)-min(impedance_X)));

disp(['NRMSE M = ',num2str(NRMSE_mag),' , ', ...
    'NRMSE ph = ',num2str(RMSE_phase),' , ',...
    'NRMSE R = ',num2str(NRMSE_real),' , ', ...
    'NRMSE X = ',num2str(NRMSE_complex)]);

%% 4 plots

%fitting data
n_points = 10000;
frequencyplot = logspace(-9,9,n_points);
omegaplot = 2*pi*frequencyplot;
Realfit = fit(1)+ (fit(2)./(1+(fit(2)^2)*(omegaplot.^2).*(fit(3).^2))) + (fit(4)./(1+(fit(4)^2)*(omegaplot.^2).*(fit(5).^2)));
Complfit = omegaplot.*fit(6)-((fit(2)^2)*omegaplot.*fit(3))./(1+(fit(2)^2)*(omegaplot.^2).*(fit(3).^2))-((fit(4)^2)*omegaplot.*fit(5))./(1+(fit(4)^2)*(omegaplot.^2).*(fit(5).^2)); 
Magfit = sqrt(Realfit.^2+Complfit.^2);
Anglefit = atand(Complfit./Realfit);

%create plot
plotstyle_data = '-';
plotstyle_fit = '--';
marker_size = 4;
line_width1 = 1.5;
line_width2 = 3;
color_data = "black"; 
color_fit = "red";  
freq_min = f_min;
freq_max = f_max;

close all
figure
tiledlayout(1,4)

nexttile
loglog(frequency,impedance_R,plotstyle_data,'LineWidth',line_width1,'Color',color_data)
hold on 
loglog(frequencyplot,Realfit,plotstyle_fit,'LineWidth',line_width2,'Color',color_fit)
grid on
legend('Data','Fit','Location','southwest')
set(gca,'FontSize',12)
xlabel('Frequency (Hz)','FontSize',14) 
ylabel('Z''(\Omega)','FontSize',14)
xlim([freq_min freq_max])
xticks([10^-6 10^0 10^6])
xticklabels({'10^{-6}','0','10^6'})
%ylim([min(impedance_R)/1.5 max(impedance_R)*1.5])

nexttile
semilogx(frequency,impedance_X,plotstyle_data,'LineWidth',line_width1,'Color',color_data)
hold on 
semilogx(frequencyplot,Complfit,plotstyle_fit,'LineWidth',line_width2,'Color',color_fit)
grid on
legend('Data','Fit','Location','southwest')
set(gca,'FontSize',12)
xlabel('Frequency (Hz)','FontSize',14) 
ylabel('Z''''(\Omega)','FontSize',14)
xlim([freq_min freq_max])
xticks([10^-6 10^0 10^6])
xticklabels({'10^{-6}','0','10^6'})
factor_im = max(impedance_X)-min(impedance_X);
%ylim([min(impedance_X)-0.1*factor_im max(impedance_X)+0.1*factor_im])
%ylim([-0.15 0.25])

nexttile
loglog(frequency,impedance_Mag,plotstyle_data,'LineWidth',line_width1,'Color',color_data)
hold on 
loglog(frequencyplot,Magfit,plotstyle_fit,'LineWidth',line_width2,'Color',color_fit)
grid on
legend('Data','Fit','Location','southwest')
set(gca,'FontSize',12)
xlabel('Frequency (Hz)','FontSize',14) 
ylabel('Magnitude (\Omega)','FontSize',14)
xlim([freq_min freq_max])
xticks([10^-6 10^0 10^6])
xticklabels({'10^{-6}','0','10^6'})
%ylim([min(impedance_Mag)/1.5 max(impedance_Mag)*1.5])


nexttile
semilogx(frequency,impedance_phase,plotstyle_data,'LineWidth',line_width1,'Color',color_data)
hold on 
semilogx(frequencyplot,Anglefit,plotstyle_fit,'LineWidth',line_width2,'Color',color_fit)
grid on
legend('Data','Fit','Location','northwest')
set(gca,'FontSize',12)
xlabel('Frequency (Hz)','FontSize',14) 
ylabel('Phase (degrees)','FontSize',14)
xlim([freq_min freq_max])
%xticks([10^-6 10^0 10^6])
%xticklabels({'10^{-6}','0','10^6'})

set(gcf,'position',[50,150,1500,350])



